##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2013 Uwe Hermann <uwe@hermann-uwe.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
##

import sigrokdecode as srd
import os
import codecs

'''
OUTPUT_PYTHON format:

Packet:
[<ptype>, <pdata>]

<ptype>, <pdata>
 - 'ITEM', [<item>, <itembitsize>]
 - 'WORD', [<word>, <wordbitsize>, <worditemcount>]

<item>:
 - A single item (a number). It can be of arbitrary size. The max. number
   of bits in this item is specified in <itembitsize>.

<itembitsize>:
 - The size of an item (in bits). For a 4-bit parallel bus this is 4,
   for a 16-bit parallel bus this is 16, and so on.

<word>:
 - A single word (a number). It can be of arbitrary size. The max. number
   of bits in this word is specified in <wordbitsize>. The (exact) number
   of items in this word is specified in <worditemcount>.

<wordbitsize>:
 - The size of a word (in bits). For a 2-item word with 8-bit items
   <wordbitsize> is 16, for a 3-item word with 4-bit items <wordbitsize>
   is 12, and so on.

<worditemcount>:
 - The size of a word (in number of items). For a 4-item word (no matter
   how many bits each item consists of) <worditemcount> is 4, for a 7-item
   word <worditemcount> is 7, and so on.
'''

def channel_list(num_channels):
    l = [{'id': 'clk', 'name': 'CLK', 'desc': 'Clock line'}]
    for i in range(num_channels):
        d = {'id': 'd%d' % i, 'name': 'D%d' % i, 'desc': 'Data line %d' % i}
        l.append(d)
    l2 = [{'id': 'cle', 'name': 'CLE', 'desc': 'CLE line'},
            {'id': 'ale', 'name': 'ALE', 'desc': 'ALE line'},
            {'id': 'we', 'name': 'WE', 'desc': 'WE line'},
            {'id': 're', 'name': 'RE', 'desc': 'RE line'}]
    l.extend(l2)
    return tuple(l)

FILENAME = "/tmp/firmware.bin"

FILESIZE = 35651584

COUNT = 8192

class Decoder(srd.Decoder):
    api_version = 2
    id = 'parallel'
    name = 'Parallel'
    longname = 'Parallel sync bus'
    desc = 'Generic parallel synchronous bus.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = ['parallel']
    optional_channels = channel_list(8)
    options = (
        {'id': 'clock_edge', 'desc': 'Clock edge to sample on',
            'default': 'rising', 'values': ('rising', 'falling')},
        {'id': 'wordsize', 'desc': 'Data wordsize', 'default': 1},
        {'id': 'endianness', 'desc': 'Data endianness',
            'default': 'little', 'values': ('little', 'big')},
    )
    annotations = (
        ('items', 'Items'),
        ('words', 'Words'),
    )

    def __init__(self):
        self.oldclk = None
        self.items = []
        self.itemcount = 0
        self.saved_item = None
        self.samplenum = 0
        self.oldpins = None
        self.ss_item = self.es_item = None
        self.first = True
        self.state = 'IDLE'
        self.oldwe = None
        self.oldre = None
        self.writing_page = False
        self.bytesnb = 0
        self.addtab = [0] * 5
        self.addr_idx = 0
        self.address = 0
        self.count = 0
        self.reading_page = True
        self.erasing_block = False
        self.wrote_files = False

        self.filetab = [ -1 ] * FILESIZE

    def start(self):
        self.out_python = self.register(srd.OUTPUT_PYTHON)
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putpb(self, data):
        self.put(self.ss_item, self.es_item, self.out_python, data)

    def putb(self, data):
        self.put(self.ss_item, self.es_item, self.out_ann, data)

    def putpw(self, data):
        self.put(self.ss_word, self.es_word, self.out_python, data)

    def putw(self, data):
        self.put(self.ss_word, self.es_word, self.out_ann, data)

    def handle_bits(self, datapins):
        # Get the bits for this item.
        item = 0
        for i in range(8):
            item |= datapins[i] << i
        assert (item >= 0 and item <= 255), "wrong item value %d" % item
        return item

    def find_we_edge(self, datapins, wepin, clepin, alepin):
        if wepin == self.oldwe:
            return
        self.oldwe = wepin

        item = self.handle_bits(datapins)
        assert (item >= 0 and item <= 255), "wrong item value %d" % item

        if wepin == 1:
            if clepin == 1:
                self.reading_page = (item == 0x90)
                if not self.erasing_block and item == 0x60:
                    self.erasing_block = True
                elif self.erasing_block and item == 0xD0:
                    self.erasing_block = False
                elif self.erasing_block:
                    assert (False), "wrong command %X while erasing" % item
                if not self.writing_page and item == 0x80:
                    self.writing_page = True
                    self.addr_idx = 0
                    self.count = 0
                elif self.writing_page and item == 0x10:
                    self.writing_page = False
                    assert (self.count == COUNT), "wrong count %d" % self.count
                elif self.writing_page:
                    assert (False), "wrong command %X" % item
                else:
                    print('CMND %X' % item, flush=True)
            elif alepin == 1:



                if self.writing_page:
                    self.addtab[self.addr_idx] = item
                    self.addr_idx += 1
                    if self.addr_idx == 1:
                        assert (item == 0), "item 0 should be 0"
                    elif self.addr_idx == 2:
                        assert (item == 0), "item 1 should be 0"
                    elif self.addr_idx == 5:
                        self.address = (self.addtab[2] + self.addtab[3] * 256 + self.addtab[4] * 256 * 256) * COUNT
                    elif self.addr_idx > 5:
                        assert (False), 'wrong value addr_idx %d' % self.addr_idx
                        print('ADDR %X' % self.address, flush=True)
            elif self.writing_page:
                assert (self.filetab[self.address] == -1), "already a value in filetab %d" % self.filetab[self.address]
                self.filetab[self.address] = item
                self.address += 1
                self.bytesnb += 1
                self.count += 1

    def find_re_edge(self, datapins, repin, clepin, alepin):
        if repin == self.oldre:
            return
        self.oldre = repin

        item = self.handle_bits(datapins)
        assert (item >= 0 and item <= 255), "wrong item value %d" % item

        if repin == 1:
            if clepin == 1:
                if self.reading_page:
                    assert (False), "wrong command %X" % item
                print('CMND %X' % item, flush=True)
            elif alepin == 1:
                if self.reading_page:
                    print('ADDR %X' % item, flush=True)
            elif self.reading_page:
                byte = '%02X' % item
                char = codecs.decode(byte, 'hex')
                print('BYTE %s (%s)' % (byte, char), flush=True)

    def decode(self, ss, es, data):
        if self.wrote_files:
            return

        for (self.samplenum, pins) in data:
            # Ignore identical samples early on (for performance reasons).
            if self.oldpins == pins:
                continue
            self.oldpins = pins

            datapins = pins[1:9]
            clepin = pins[9]
            alepin = pins[10]
            wepin = pins[11]
            repin = pins[12]

            # State machine.
            if self.state == 'IDLE':
                if pins[0] not in (0, 1):
                    self.find_we_edge(datapins, wepin, clepin, alepin)
                    self.find_re_edge(datapins, repin, clepin, alepin)
                else:
                    raise Exception('Invalid clock !')
            else:
                raise Exception('Invalid state: %s' % self.state)
        if self.bytesnb > 35651500:
            print ('NBBYTES = %d, COUNT = %d' % (self.bytesnb, self.count))
            if self.filetab.count(-1) == 0:
                other = open(FILENAME, "wb")
                other.write(bytes(self.filetab))
                other.flush()
                os.fsync(other.fileno())
                self.wrote_files = True
